
#include "browser.h"

//////////////////////////////////////////////////////////////////////
// IUnknown implementation
//////////////////////////////////////////////////////////////////////

HRESULT STDMETHODCALLTYPE Browser::QueryInterface(REFIID riid,void __RPC_FAR *__RPC_FAR *ppvObject)
{
	*ppvObject = 0;
	if (IID_IUnknown == riid)			{ *ppvObject = (IUnknown*)(IOleClientSite*)this; }
	if (IID_IOleInPlaceSite == riid)	{ *ppvObject = (IOleInPlaceSite*)this; }
	if (DIID_DWebBrowserEvents2 == riid){ *ppvObject = (DWebBrowserEvents2*)this; }
	if (*ppvObject)
		return S_OK;
	return E_NOINTERFACE;
}
ULONG STDMETHODCALLTYPE Browser::AddRef(void)	{ return 1; }
ULONG STDMETHODCALLTYPE Browser::Release(void)	{ return 1; }

//////////////////////////////////////////////////////////////////////
// IOleWindow implementation
//////////////////////////////////////////////////////////////////////

HRESULT STDMETHODCALLTYPE Browser::GetWindow(HWND __RPC_FAR *phwnd)
{
	*phwnd = parent;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE Browser::ContextSensitiveHelp(BOOL)	{ return E_NOTIMPL; }

//////////////////////////////////////////////////////////////////////
// IOleClientSite implementation
//////////////////////////////////////////////////////////////////////

HRESULT STDMETHODCALLTYPE Browser::SaveObject(void)				{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::GetMoniker(DWORD,DWORD,IMoniker __RPC_FAR *__RPC_FAR *)	{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::GetContainer(IOleContainer __RPC_FAR *__RPC_FAR *)		{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::ShowObject(void)				{ return S_OK; }
HRESULT STDMETHODCALLTYPE Browser::OnShowWindow(BOOL)			{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::RequestNewObjectLayout(void) { return E_NOTIMPL; }

//////////////////////////////////////////////////////////////////////
// IOleInPlaceSite implementation
//////////////////////////////////////////////////////////////////////

HRESULT STDMETHODCALLTYPE Browser::GetWindowContext(IOleInPlaceFrame __RPC_FAR *__RPC_FAR *ppFrame,IOleInPlaceUIWindow __RPC_FAR *__RPC_FAR *ppDoc,LPRECT lprcPosRect,LPRECT lprcClipRect,LPOLEINPLACEFRAMEINFO lpFrameInfo)
{
	*ppFrame	= 0;
	*ppDoc		= 0;

	//adjust window size
	GetClientRect(parent,lprcPosRect);
	lprcPosRect->bottom += 2; lprcPosRect->left -= 2; lprcPosRect->right += 2; lprcPosRect->top -= 2;
	*lprcClipRect = *lprcPosRect;

	lpFrameInfo	= 0;
	return S_OK;
}
HRESULT STDMETHODCALLTYPE Browser::Scroll(SIZE)				{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::OnUIDeactivate(BOOL)		{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::OnInPlaceDeactivate()	{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::DiscardUndoState()		{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::DeactivateAndUndo()		{ return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::OnPosRectChange(LPCRECT)	{ return S_OK; }
HRESULT STDMETHODCALLTYPE Browser::OnInPlaceActivate()		{ return S_OK; }
HRESULT STDMETHODCALLTYPE Browser::OnUIActivate()			{ return S_OK; }
HRESULT STDMETHODCALLTYPE Browser::CanInPlaceActivate()		{ return S_OK; }

//////////////////////////////////////////////////////////////////////
// IDispatch implementation
//////////////////////////////////////////////////////////////////////

HRESULT STDMETHODCALLTYPE Browser::GetTypeInfoCount(UINT __RPC_FAR *pctinfo)
{
	*pctinfo = 0;
	return S_OK;
}
HRESULT STDMETHODCALLTYPE Browser::GetTypeInfo(UINT,LCID,ITypeInfo __RPC_FAR *__RPC_FAR*) { return E_NOTIMPL; }
HRESULT STDMETHODCALLTYPE Browser::GetIDsOfNames(REFIID,LPOLESTR __RPC_FAR*,UINT,LCID,DISPID __RPC_FAR*) { return E_NOTIMPL; }

//////////////////////////////////////////////////////////////////////
// Attaching/Detaching
//////////////////////////////////////////////////////////////////////

BOOL Browser::attach(HWND hParent,WNDPROC new_wnd_proc)
{
	RECT rc;
	GetClientRect(hParent,&rc);
	parent = hParent;
	if(
		//request for IWebBrowser2
		SUCCEEDED(CoCreateInstance(CLSID_WebBrowser,0,CLSCTX_INPROC_SERVER,IID_IWebBrowser2,(LPVOID*)&pWebBrowser2))
		&&
		//request for IOleObject
		SUCCEEDED(pWebBrowser2->QueryInterface(IID_IOleObject,(LPVOID*)&pOleObject))
		&&
		//request for IOleInPlaceObject
		SUCCEEDED(pWebBrowser2->QueryInterface(IID_IOleInPlaceObject,(LPVOID*)&pOleInPlaceObject))
		&&
		//request for IConnectionPointContainer
		SUCCEEDED(pWebBrowser2->QueryInterface(IID_IConnectionPointContainer,(LPVOID*)&pCPC))
		&&
		//request for IConnectionPoint (DWebBrowserEvents2)
		SUCCEEDED(pCPC->FindConnectionPoint(DIID_DWebBrowserEvents2,&pCP))
		&&
		//advise sink
		SUCCEEDED(pCP->Advise((IUnknown*)(IOleClientSite*)this,&cookie))
		&&
		//client site
		SUCCEEDED(pOleObject->SetClientSite((IOleClientSite*)this))
		&&
		//in place activation
		SUCCEEDED(pOleObject->DoVerb(OLEIVERB_INPLACEACTIVATE,0,(IOleClientSite*)this,0,parent,&rc))
		)
	{
		old_wnd_proc = (WNDPROC)SetWindowLong(parent,GWL_WNDPROC,(LONG)new_wnd_proc);
		return TRUE;
	}
	return FALSE;
}

void Browser::detach()
{
	//releasing interface pointers
	if (pCP) { pCP->Unadvise(cookie); pCP->Release(); pCP = 0; }
	if (pCPC) { pCPC->Release(); pCPC = 0; }
	if (pOleInPlaceObject){ pOleInPlaceObject->Release(); pOleInPlaceObject = 0; }
	if (pOleObject) { pOleObject->Close(OLECLOSE_NOSAVE); pOleObject->Release(); pOleObject = 0; }
	if (pWebBrowser2) { pWebBrowser2->Release(); pWebBrowser2 = 0; }
	
	//unsubclassing
	if (old_wnd_proc)
	{
		SetWindowLong(parent,GWL_WNDPROC,(LONG)old_wnd_proc);
		old_wnd_proc = 0;
	}
	parent = 0;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Browser::Browser() : 
	parent				(0),
	pWebBrowser2		(0),
	pOleObject			(0),
	pOleInPlaceObject	(0),
	pCPC				(0),
	pCP					(0),
	old_wnd_proc		(0)
{}
Browser::~Browser() { detach(); }