//------------------------------------------------------------------------------------------
// nHTMLn.dll - HTML renderer for mIRC
// 06.05.2001 (c) Necroman - necroman@europe.com, #mIRC @ Undernet
//------------------------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <exdispid.h>
#include "Browser.h"
#include "MircMap.h"

////////////////////////////////////////////////////////////////////////////////////////////
// nHTMLn declarations
////////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
	DWORD  mVersion;
	HWND   mHwnd;
	BOOL   mKeep;
} LOADINFO;

#define OBJ_LIMIT 256

const LPCSTR E_INVALID_WINDOW		= "E_INVALID_WINDOW";
const LPCSTR E_FAILED				= "E_FAILED";
const LPCSTR E_TOO_MANY_BROWSERS	= "E_TOO_MANY_BROWSERS";
const LPCSTR E_NOT_FOUND			= "E_NOT_FOUND";
const LPCSTR E_ALREADY_ATTACHED		= "E_ALREADY_ATACHED";
const LPCSTR SCANCEL				= "S_CANCEL";
const LPCSTR SOK					= "S_OK";

class MircBrowser : public Browser
{
	HRESULT STDMETHODCALLTYPE Invoke(DISPID,REFIID,LCID,WORD,DISPPARAMS __RPC_FAR *,VARIANT __RPC_FAR *,EXCEPINFO __RPC_FAR *,UINT __RPC_FAR *);	
	VARIANT_BOOL dispatch(LPSTR event,BSTR str);
};

MircMap			mircmap;
MircBrowser		*browsers[OBJ_LIMIT];
HWND			sel_wnd;


////////////////////////////////////////////////////////////////////////////////////////////
// nHTMLn internal functions
////////////////////////////////////////////////////////////////////////////////////////////

#define ret(x) { return ret_mirc(data,x); } 
inline int ret_mirc(char *data,const char *text)
{
	lstrcpy(data,text);
	return 3;
}

//------------------------------------------------------------------------------------------
//find browser object by HWND
int get_browser(HWND parent)
{
	for (int i = OBJ_LIMIT; --i;)
		if (browsers[i] && browsers[i]->parent == parent)
			return i;
	return 0;
}

//------------------------------------------------------------------------------------------
//find browser object by HWND
int get_browser(LPSTR data)
{
	HWND window = (HWND)atol(data);
	if (!IsWindow(window)) window = sel_wnd;
	return get_browser(window);	
}

//------------------------------------------------------------------------------------------
//subclass window procedure
LRESULT CALLBACK subclass_wnd_proc(HWND hwnd,UINT msg,WPARAM wp,LPARAM lp)
{
	static RECT rc;
	static int i;
	
	//find attached browser
	i = get_browser(hwnd);
	if (i)
	{
		//process special messages
		WNDPROC old_proc = browsers[i]->old_wnd_proc;
		switch(msg)
		{
		case WM_SIZE:	//adjust browser size
			GetClientRect(hwnd,&rc);
			rc.bottom += 2; rc.left -= 2; rc.right += 2; rc.top -= 2;
			browsers[i]->pOleInPlaceObject->SetObjectRects(&rc,&rc);
			break;
		case WM_DESTROY:
			delete browsers[i];	//detach browser
			browsers[i] = 0;
			break;
		}
		return CallWindowProc(old_proc,hwnd,msg,wp,lp);
	}
	return 0; //should never be here
}

//------------------------------------------------------------------------------------------
//search for child windows
char findstr[MAX_PATH];
BOOL CALLBACK enum_children(HWND hWnd,LPARAM lp)
{
	char caption[MAX_PATH];
	GetWindowText(hWnd,caption,MAX_PATH);
	if (!strcmp(findstr,caption))
	{
		wsprintf((char*)lp,"%d",hWnd);
		return FALSE;
	}
	return TRUE;
}

//------------------------------------------------------------------------------------------
//dispatch nHTMLn event
VARIANT_BOOL MircBrowser::dispatch(LPSTR event,BSTR bstr)
{
	if (*handler)
	{
		//evaluate $handler(hwnd,event [ ,bstr ] )
		String mirc('$');
		mirc <= handler <= '(' <= (int)parent <= ',' <= event;
		if (bstr && *bstr) mirc <= ',' <= bstr;
		mirc <= ')';
		LPCSTR result = mircmap.eval(mirc);

		if (!lstrcmpi(SCANCEL,result))	return VARIANT_TRUE;
		if (lstrcmpi(SOK,result))		handler = ""; 
	}
	return VARIANT_FALSE;
}

//------------------------------------------------------------------------------------------
//event sink
HRESULT MircBrowser::Invoke(DISPID dispIdMember,REFIID riid,LCID lcid,WORD wFlags,DISPPARAMS __RPC_FAR *pDispParams,VARIANT __RPC_FAR *pVarResult,EXCEPINFO __RPC_FAR *pExcepInfo,UINT __RPC_FAR *puArgErr)
{
	UINT err;
	VARIANT res = { 0 };
	DispGetParam(pDispParams,1,VT_BSTR,&res,&err);
	
	switch (dispIdMember)
	{
	case DISPID_NAVIGATECOMPLETE2:
		dispatch("navigate_complete",res.bstrVal);
		break;
	case DISPID_BEFORENAVIGATE2:
		*pDispParams->rgvarg->pboolVal = dispatch("navigate_begin",res.bstrVal);
		break;
	case DISPID_DOCUMENTCOMPLETE:
		dispatch("document_complete",res.bstrVal);
		break;    
	case DISPID_DOWNLOADBEGIN:
		dispatch("download_begin",0);
		break;
	case DISPID_DOWNLOADCOMPLETE:
		dispatch("download_complete",0);
		break;
	case DISPID_NEWWINDOW2:
		*pDispParams->rgvarg->pboolVal = dispatch("window_open",0);
		break;
	case DISPID_STATUSTEXTCHANGE:
		DispGetParam(pDispParams,0,VT_BSTR,&res,&err);
		dispatch("status_change",res.bstrVal);
		break;
	case DISPID_TITLECHANGE:
		DispGetParam(pDispParams,0,VT_BSTR,&res,&err);
		dispatch("title_change",res.bstrVal);
		break;
	case DISPID_PROGRESSCHANGE:
		{
			String str(" of ");
			str <= res.bstrVal;
			DispGetParam(pDispParams,0,VT_BSTR,&res,&err);
			OLECHAR ustr[MAX_PATH << 2];
			(String(res.bstrVal) <= str).unicode(ustr);
			dispatch("progress_change",ustr);
		}
		break;
	case DISPID_ONVISIBLE:
		{
			DispGetParam(pDispParams,0,VT_BOOL,&res,&err);
			OLECHAR ustr[MAX_PATH];
			String(res.boolVal ? "visible" : "invisible").unicode(ustr);
			dispatch("visibility_change",ustr);
		}
		break;
	case DISPID_COMMANDSTATECHANGE:
		{
			String str(':');
			str <= (res.boolVal ? "enabled" : "disabled");
			DispGetParam(pDispParams,0,VT_I4,&res,&err);
			if (res.lVal < 0) break;
			switch(res.lVal)
			{
			case 1: str = ((String)"forward" <= str); break;
			case 2: str = ((String)"back" <= str); break;
			default: str = ((String)res.lVal <= str);
			}
			OLECHAR ustr[MAX_PATH];
			str.unicode(ustr);
			dispatch("state_change",ustr);
		}
		break; 
	}
	VariantClear(&res);
	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////////////////
// nHTMLn exported functions
////////////////////////////////////////////////////////////////////////////////////////////

extern "C"
{
	//------------------------------------------------------------------------------------------
	//attach browser to window
	int __stdcall attach(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		HWND window = (HWND)atol(data);
		if (!IsWindow(window)) window = sel_wnd;
		if (IsWindow(window))
		{
			//check for exising
			int index = 0;
			for (int i = OBJ_LIMIT; --i;)
			{
				if (!browsers[i]) index = index < i ? i : index;
				else if (browsers[i]->parent == window) ret(E_ALREADY_ATTACHED);
			}
			if (index)
			{	
				MircBrowser *pb = new MircBrowser();
				if (!(pb->attach(window,subclass_wnd_proc))) { delete pb; ret(E_FAILED); }
				browsers[index] = pb;
				sel_wnd = window;
				ret(SOK);
			}
			ret(E_TOO_MANY_BROWSERS);
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//detach browser from window
	int __stdcall detach(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i)
		{
			delete browsers[i];
			browsers[i] = 0;
			ret(SOK);
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//select default browser
	int __stdcall select(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		HWND window = (HWND)atol(data);
		int i = get_browser(window);
		if (i)
		{
			sel_wnd = window;
			ret(SOK);
		}
		ret(E_INVALID_WINDOW);
	}

	//------------------------------------------------------------------------------------------
	//return HWND of default browser
	int __stdcall selected(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		wsprintf(data,"%d",sel_wnd);
		return 3;
	}
	
	//------------------------------------------------------------------------------------------
	//navigate to URL
	int __stdcall navigate(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		HWND window = sel_wnd;
		if (IsWindow(window))
		{
			OLECHAR url[MAX_PATH << 4];
			String(data).unicode(url);
			
			int i = get_browser(window);
			if (i)
			{
				VARIANT v = { 0 };			
				ret(SUCCEEDED(browsers[i]->pWebBrowser2->Navigate(url,&v,&v,&v,&v)) ? SOK : E_FAILED);
			}
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//set handler
	int __stdcall handler(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		HWND window = sel_wnd;
		if (IsWindow(window))
		{
			int i = get_browser(window);
			if (i)
			{
				browsers[i]->handler = data;
				ret(SOK);
			}
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//navigate back
	int __stdcall back(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->GoBack()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//navigate home
	int __stdcall home(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->GoHome()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//navigate forward
	int __stdcall forward(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->GoForward()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//search
	int __stdcall search(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->GoSearch()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//reload page
	int __stdcall refresh(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->Refresh()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//stop downloading page
	int __stdcall stop(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i) ret(SUCCEEDED(browsers[i]->pWebBrowser2->Stop()) ? SOK : E_FAILED);
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//location name
	int __stdcall name(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i)
		{
			BSTR str;
			if(FAILED(browsers[i]->pWebBrowser2->get_LocationName(&str))) ret(E_FAILED);
			lstrcpy(data,(String)str);
			SysFreeString(str);
			return 3;
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//location URL
	int __stdcall url(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i)
		{
			BSTR str;
			if(FAILED(browsers[i]->pWebBrowser2->get_LocationURL(&str))) ret(E_FAILED);
			lstrcpy(data,(String)str);
			SysFreeString(str);
			return 3;
		}
		ret(E_INVALID_WINDOW);
	}

	//------------------------------------------------------------------------------------------
	//find HWND from caption
	int __stdcall find(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		lstrcpy(findstr,data);
		lstrcpy(data,E_NOT_FOUND);
		EnumChildWindows(GetDesktopWindow(),enum_children,(LPARAM)data);
		return 3;
	}

	//------------------------------------------------------------------------------------------
	//get caption from HWND
	int __stdcall caption(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		HWND window = (HWND)atol(data);
		lstrcpy(data,E_NOT_FOUND);
		if (IsWindow(window))
			GetWindowText(window,data,800);
		return 3;
	}
	
	//------------------------------------------------------------------------------------------
	//print
	int __stdcall print(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		int i = get_browser(data);
		if (i)
		{
			LPDISPATCH lpDispatch = 0;
			LPOLECOMMANDTARGET lpOleCommandTarget = 0;
			LPCSTR result = E_FAILED;
			if(SUCCEEDED(browsers[i]->pWebBrowser2->get_Document(&lpDispatch)))
			{
				if(SUCCEEDED(lpDispatch->QueryInterface(IID_IOleCommandTarget,(void**)&lpOleCommandTarget)))
				{
					if(SUCCEEDED(lpOleCommandTarget->Exec(0,OLECMDID_PRINT,0,0,0)))
						result = SOK;
					lpOleCommandTarget->Release();
				}
				lpDispatch->Release();
			}
			ret(result);
		}
		ret(E_INVALID_WINDOW);
	}
	
	//------------------------------------------------------------------------------------------
	//version info
	int __stdcall version(HWND,HWND,char *data,char*,BOOL,BOOL)
	{
		ret("nHTMLn 2.5 - (c) 2001 Necroman, necroman@europe.com, #mIRC @ Undernet");
	}
	
	//------------------------------------------------------------------------------------------
	void __stdcall LoadDll(LOADINFO *li)
	{
		OleInitialize(0);
		mircmap.init(li->mHwnd);
	}

	//------------------------------------------------------------------------------------------
	int __stdcall UnloadDll(int timeout)
	{
		if (!timeout)
		{
			//destroy browsers
			for (int i = OBJ_LIMIT; i--;)
				if (browsers[i]) delete browsers[i];

			OleUninitialize();
		}
		return 0;
	}
	
	//------------------------------------------------------------------------------------------
	// :-)
	int __stdcall egg(HWND mWnd,HWND,char *data,char*,BOOL,BOOL)
	{
		SetWindowText(mWnd,lstrcpy(data,"All your base are belong to us"));
		return 3; 
	}
}